import time
import matplotlib.pyplot as plt
import numpy as np
import sys
import math
from functools import reduce

# Increase recursion limit for testing larger numbers
sys.setrecursionlimit(3000)

# Factorial Implementations (7 Methods)
def factorial_loop(n):
    result = 1
    for i in range(1, n + 1):
        result *= i
    return result

def factorial_recursion(n):
    if n == 0 or n == 1:
        return 1
    return n * factorial_recursion(n - 1)

def factorial_memoization(n, memo={}):
    if n in memo:
        return memo[n]
    if n == 0 or n == 1:
        return 1
    memo[n] = n * factorial_memoization(n - 1, memo)
    return memo[n]

def factorial_builtin(n):
    return math.factorial(n)

def factorial_reduce(n):
    return reduce(lambda x, y: x * y, range(1, n + 1), 1)

def factorial_numpy(n):
    import numpy as np
    return np.prod(np.arange(1, n + 1))

def factorial_while(n):
    result = 1
    i = 1
    while i <= n:
        result *= i
        i += 1
    return result

# Measure Execution Time with Multiple Runs
def measure_time(func, n, runs=10):
    times = []
    for _ in range(runs):
        start = time.perf_counter()
        func(n)
        end = time.perf_counter()
        times.append((end - start) * 1000)  # Convert to milliseconds
    return sum(times) / len(times)  # Return average execution time

# Define test values
test_values = np.arange(1, 1001, 100)  # Testing for n from 1 to 1000

# Store execution times (averaged over 10 runs)
times_loop = [measure_time(factorial_loop, n) for n in test_values]
times_recursion = [measure_time(factorial_recursion, n) for n in test_values if n < 1000]  # Avoid stack overflow
times_memoization = [measure_time(factorial_memoization, n) for n in test_values]
times_builtin = [measure_time(factorial_builtin, n) for n in test_values]
times_reduce = [measure_time(factorial_reduce, n) for n in test_values]
times_numpy = [measure_time(factorial_numpy, n) for n in test_values]
times_while = [measure_time(factorial_while, n) for n in test_values]

# Plot Execution Time Graph (Benchmark)
plt.figure(figsize=(10, 6))
plt.plot(test_values, times_loop, label="Loop", marker="o", linestyle="dashed")
plt.plot(test_values[:len(times_recursion)], times_recursion, label="Recursion", marker="x", linestyle="dashed")
plt.plot(test_values, times_memoization, label="Memoization", marker="s", linestyle="dashed")
plt.plot(test_values, times_builtin, label="math.factorial()", marker="^", linestyle="dashed")
plt.plot(test_values, times_reduce, label="Reduce", marker="d", linestyle="dashed")
plt.plot(test_values, times_numpy, label="NumPy", marker="p", linestyle="dashed")
plt.plot(test_values, times_while, label="While Loop", marker="*", linestyle="dashed")

plt.xlabel("n (Factorial Input)")
plt.ylabel("Execution Time (ms)")
plt.title("Factorial Computation Time Benchmark (7 Methods)")
plt.legend()
plt.grid(True)

# Save as JPG instead of PNG
execution_time_graph_jpg = "factorial_execution_time.jpg"
plt.savefig(execution_time_graph_jpg, format="jpg", dpi=300)
print(f"Graph saved as {execution_time_graph_jpg}")
